"""
This is a package that allows you to perform calculations involving quantities with known or estimated uncertainties.

It exposes the class Measurement, which is a numeric-like class that can be used with common arithmetic operations.

This package also includes a way to parse files (through create_from_file), giving the corresponding mean and sigma.
"""


class Measurement:
    """
    This class is for representing measurements
    in the form of (VAL +- UNC), where the VAL is the mean value
    and UNC is the uncertainty of the given value.

    It supports addition, subtraction, division and multiplication
    between Measurements(or with regular floating-point numbers).
    """
    def __init__(self, val: float, unc: float):
        """
        To instantiate a Measurement, you need:

        :param val: the mean value
        :param unc: the uncertainty
        """
        self.val = val
        self.unc = unc

    def __add__(self, m2):
        # TODO: ADD DOCUMENTATION
        val = self.val + m2.val
        unc = norm(self.unc, m2.unc)
        return Measurement(val, unc)

    def __sub__(self, m2):
        # TODO: ADD DOCUMENTATION
        val = self.val - m2.val
        unc = norm(self.unc, m2.unc)
        return Measurement(val, unc)

    def __mul__(self, m2):
        # TODO: ADD DOCUMENTATION
        if isinstance(m2, Measurement):
            val = self.val * m2.val
            unc = val * norm(self.unc / self.val, m2.unc / m2.val)
            return Measurement(val, unc)
        # Otherwise, it's a scalar
        else:
            return Measurement(self.val * m2, self.unc * m2)

    # Define multiplication as commutative
    __rmul__ = __mul__

    def __truediv__(self, m2):
        # TODO: ADD DOCUMENTATION
        if isinstance(m2, Measurement):
            val = self.val / m2.val
            unc = val * norm(self.unc / self.val, m2.unc / m2.val)
            return Measurement(val, unc)
        else:
            return Measurement(self.val / m2, self.unc / m2)

    def __pow__(self, exp):
        # TODO: ADD DOCUMENTATION
        return Measurement(self.val ** exp, exp * self.val ** (exp - 1) * self.unc)

    def __str__(self):
        # TODO: improve this printing
        # TODO: Figure out how to use localization(E.g. US vs European notation, etc)
        return f"{self.val:.3f} \\pm {self.unc:.3f}"


def norm(*x: float) -> float:
    """Returns the norm of x, i.e. the squareroot of the sum of squares in the tuple x"""
    return (sum([i ** 2 for i in x])) ** 0.5



from typing import List

from . import measurement as m


def read(filename: str, headerline=False, col=0, totalcol=1) -> List[float]:
    """
    Reads the data from the file.

    :param filename: the relative/absolute path of the file in your system
    :param headerline: whether or not there is a header in your file
    :param col: TODO: ADD DOCUMENTATION
    :param totalcol: TODO: ADD DOCUMENTATION
    :return: returns the data as a list of floating-point numbers
    """
    # Reads the data from the file.
    # The file MUST have only the raw data
    with open(filename, 'r') as f:
        s = f.read()
        if headerline:
            s = s.split('\n', 1)[1]
        # For European/Brazilian number notation
        s = s.replace(",", ".")
        data = [float(val) for i, val in enumerate(s.split()) if i % totalcol == col]
    return data


def sigma_m(data: List[float], unc: float = 0):
    """
    Calculates the standard deviation of the mean.

    Rationale: just calculating the standard deviation of the
    data isn't enough, we must also account for the estimated
    uncertainty 'unc' of the experiment.

    :param data: the data of the experiment, as a list of floats
    :param unc: the estimated uncertainty of the experiment
    :return: the standard deviation of the mean
    """
    # Data must be number array, for use with unc, or
    length = len(data)
    av = sum(data) / length
    # deviation for the distribution
    dist_sigma = (sum([(x - av) ** 2 for x in data]) / (length - 1)) ** 0.5
    # compounding measurement uncertainty with distribution uncertainty
    total_sigma = m.norm(dist_sigma, unc)
    return total_sigma / length ** 0.5


def create(filename: str, headerline=False, col=0, totalcol=1, unc: float = 0):
    # TODO: ADD DOCUMENTATION
    data = read(filename, headerline, col, totalcol)
    return m.Measurement(sum(data) / len(data), sigma_m(data, unc))
